/**************************************************************************
 *                                                                        *
 * Copyright (C) 2016 Malte Brandy <malte.brandy@maralorn.de>                        *
 *                                                                        *
 * This program is free software; you can redistribute it and/or          *
 * modify it under the terms of the GNU General Public License            *
 * as published by the Free Software Foundation; either version 3         *
 * of the License, or (at your option) any later version.                 *
 *                                                                        *
 * This program is distributed in the hope that it will be useful,        *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.  *
 *                                                                        *
 **************************************************************************/

#pragma once

#include <QtWidgets/QApplication>

class MainWindow;

class ActivityDetector : public QObject
{
    Q_OBJECT

    public:
        ActivityDetector(QApplication& a, MainWindow& w);

    public slots:
        void updateEnabled();
        void setEnabled(bool enabled);

    signals:
        void triggered();

    protected:
        bool eventFilter(QObject* obj, QEvent* ev) override;

    private:
        QApplication& m_app;
        MainWindow& m_mainWindow;
        bool m_enabled;
};
